/**
 * \copyright Copyright (C) 2019 Texas Instruments Incorporated - http://www.ti.com/
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "ads131m0x.h"
//extern int32_t fd;

#define ADC_RETRY_CNT (3U)
#define ADC_SPI_DEVICE_NODE ("/dev/spidev0.0")
#define ADC_SPI_COMM_MODE (2U)
#define ADC_SPI_COMM_BITS (8U)             /* Data Size */
//#define ADC_SPI_COMM_SPEED (10000000U)     /* 10MHz */
#define ADC_SPI_COMM_SPEED (8192000U)    


#define SYNC_RESET_DEVICE     ("5")
#define SYNC_RESET_OFFSETS     ("15")    

#define DRDY_DEVICE     ("5") // BANK
#define DRDY_OFFSETS     ("12")    // PIN


static const char *device = ADC_SPI_DEVICE_NODE;

int32_t fd = 0;
static uint32_t mode = (uint32_t)ADC_SPI_COMM_MODE;
static uint8_t bits = ADC_SPI_COMM_BITS;
static uint32_t speed = ADC_SPI_COMM_SPEED;

int interrupt_flag = 0;

//****************************************************************************
//
// Internal variables
//
//****************************************************************************

/* Do we need to handle this flag like below or it can handle like in gpio drdy pin */
/* gpio mon utility need to use */


// Flag to indicate if a /DRDY interrupt has occurred
static volatile bool flag_nDRDY_INTERRUPT = false;



//****************************************************************************
//
// Internal function prototypes
//
//****************************************************************************
void InitGPIO(void);
void InitSPI(void);
void GPIO_DRDY_IRQHandler(void);

// ADC required 
int syncResetGPIOOp(int value);
int DRDYGPIOop();
void func (char * string);
int make_signalfd(void);
const char *get_progname(void);

FILE* pFile;
FILE* channel0ptr;
FILE* channel1ptr;

//****************************************************************************
//
// Internal variables
//
//****************************************************************************

// Array used to recall device register map configurations */
static uint16_t             registerMap[NUM_REGISTERS];

// Array of SPI word lengths
const static uint8_t        wlength_byte_values[] = {2, 3, 4, 4};



//****************************************************************************
//
// Internal function prototypes
//
//****************************************************************************

uint8_t     buildSPIarray(const uint16_t opcodeArray[], uint8_t numberOpcodes, uint8_t byteArray[]);
uint16_t    enforce_selected_device_modes(uint16_t data);
uint8_t     getWordByteLength(void);


//****************************************************************************
//
// External Functions (prototypes declared in hal.h)
//
//****************************************************************************


//*****************************************************************************
//
//! Initializes MCU peripherals for interfacing with the ADC.
//!
//! \fn void InitADC(void)
//!
//! \return None.
//
//*****************************************************************************
void main(void)
{
   pFile = fopen("log.txt","wb");
   unsigned int num_lines = 0;

   struct timespec timeout = {10, 0};
    // IMPORTANT: Make sure device is powered before setting GPIOs pins to HIGH state.
    //printf("ads131m02: Init GPIO started \n");
    // Initialize GPIOs pins used by ADS131M0x
    InitGPIO();
    //printf("ads131m02: Init GPIO completed \n");

    //printf("ads131m02: Init SPI started \n");
    // Initialize SPI peripheral used by ADS131M0x
    InitSPI();
    //printf("ads131m02: Init SPI completed \n");
    adcStartup();
    //printf("ads131m02: ADC startup completed \n");
   
   /* 
    while(1) {
        sleep(2);
    }*/

}

void func (char * string)
{
    char buffer[100];
    strcpy(buffer, "log print \n");
    fwrite(buffer,1,sizeof(buffer),pFile);
}



struct mon_ctx {
        unsigned int offset;
        unsigned int events_wanted;
        unsigned int events_done;

        bool silent;
        char *fmt;

        int sigfd;
};


struct callback_data {
        /* Replace with a union once we have more modes using callback data. */
        struct timeval tv;
        bool daemonize;
};

static void maybe_daemonize(bool daemonize)
{
        int rv;

        if (daemonize) {
                rv = daemon(0, 0);
		printf("ads131m02: may be daemonize \n");
                if (rv < 0)
                        die("unable to daemonize: %s", strerror(errno));
        }
}

static void wait_enter(void *data GPIOD_UNUSED)
{
        getchar();
}

static void wait_time(void *data)
{
        struct callback_data *cbdata = data;

        maybe_daemonize(cbdata->daemonize);
        select(0, NULL, NULL, NULL, &cbdata->tv);
}

static void wait_signal(void *data)
{
        struct callback_data *cbdata = data;
        struct pollfd pfd;
        int sigfd, rv;

        sigfd = make_signalfd();

        memset(&pfd, 0, sizeof(pfd));
        pfd.fd = sigfd;
        pfd.events = POLLIN | POLLPRI;

        maybe_daemonize(cbdata->daemonize);
	printf("ads131m02: wait signal \n");
        for (;;) {
                rv = poll(&pfd, 1, 1000 /* one second */);
                if (rv < 0)
                        die("error polling for signals: %s", strerror(errno));
                else if (rv > 0)
                        break;
        }

	printf("ads131m02: ADC reading signal info \n");

        /*
         * Don't bother reading siginfo - it's enough to know that we
         * received any signal.
         */
        close(sigfd);
}


enum {
        MODE_EXIT = 0,
        MODE_WAIT,
        MODE_TIME,
        MODE_SIGNAL,
};

struct mode_mapping {
        int id;
        const char *name;
        gpiod_ctxless_set_value_cb callback;
};

static const struct mode_mapping modes[] = {
        [MODE_EXIT] = {
                .id             = MODE_EXIT,
                .name           = "exit",
                .callback       = NULL,
        },
        [MODE_WAIT] = {
                .id             = MODE_WAIT,
                .name           = "wait",
                .callback       = wait_enter,
        },
        [MODE_TIME] = {
                .id             = MODE_TIME,
                .name           = "time",
                .callback       = wait_time,
        },
        [MODE_SIGNAL] = {
                .id             = MODE_SIGNAL,
                .name           = "signal",
                .callback       = wait_signal,
        },
};

static const struct mode_mapping *parse_mode(const char *mode)
{
        unsigned int i;

	printf("ads131m02: mode mapping \n ");
        for (i = 0; i < ARRAY_SIZE(modes); i++)
                if (strcmp(mode, modes[i].name) == 0)
                        return &modes[i];

        return NULL;
}




//****************************************************************************
//
// Timing functions
//
//****************************************************************************



//*****************************************************************************
//
//! Provides a timing delay with 'ms' resolution.
//!
//! \fn void delay_ms(const uint32_t delay_time_ms)
//!
//! \param delay_time_ms is the number of milliseconds to delay.
//!
//! \return None.
//
//*****************************************************************************

/* Is it different for the processor or we can go with this delay_ms function */

void delay_ms(const uint32_t delay_time_ms)
{
    /* --- INSERT YOUR CODE HERE --- */
    usleep(delay_time_ms*1000);
    /* const uint32_t cycles_per_loop = 3;
    MAP_SysCtlDelay( delay_time_ms * getSysClockHz() / (cycles_per_loop * 1000u) );
    */
}


/*****************************************************************************
//
//! Provides a timing delay with 'us' resolution.
//!
//! \fn void delay_us(const uint32_t delay_time_us)
//!
//! \param delay_time_us is the number of microseconds to delay.
//!
//! \return None.
//
//*****************************************************************************

/* Is it different for the processor or we can go with this delay_us function */

void delay_us(const uint32_t delay_time_us)
{
    /* --- INSERT YOUR CODE HERE --- */
    usleep(delay_time_us);
    /*const uint32_t cycles_per_loop = 3;
    MAP_SysCtlDelay( delay_time_us * getSysClockHz() / (cycles_per_loop * 1000000u) );
    */
}

static void event_print_custom(unsigned int offset,
			       const struct timespec *ts,
			       int event_type,
			       struct mon_ctx *ctx)
{
	char *prev, *curr, fmt;

	for (prev = curr = ctx->fmt;;) {
		curr = strchr(curr, '%');
		if (!curr) {
			fputs(prev, stdout);
			break;
		}

		if (prev != curr)
			fwrite(prev, curr - prev, 1, stdout);

		fmt = *(curr + 1);

		switch (fmt) {
		case 'o':
			printf("%u", offset);
			break;
		case 'e':
			if (event_type == GPIOD_CTXLESS_EVENT_CB_RISING_EDGE)
				fputc('1', stdout);
			else
				fputc('0', stdout);
			break;
		case 's':
			printf("%ld", ts->tv_sec);
			break;
		case 'n':
			printf("%ld", ts->tv_nsec);
			break;
		case '%':
			fputc('%', stdout);
			break;
		case '\0':
			fputc('%', stdout);
			goto end;
		default:
			fwrite(curr, 2, 1, stdout);
			break;
		}

		curr += 2;
		prev = curr;
	}

end:
	fputc('\n', stdout);
}

static void event_print_human_readable(unsigned int offset, const struct timespec *ts, int event_type)
 {
         char *evname;
	 printf("ads131m02: human readable event print \n");

         if(event_type == GPIOD_CTXLESS_EVENT_CB_RISING_EDGE)
         {
		 printf("ads131m02: RISING EDGE \n");
                 evname = "RISING EDGE";
         }
         else
         {
		 printf("ads131m02: ADC FALLING EDGE \n");
                 evname = "FALLING EDGE";
         }

         printf("event: %s offset: %u timestamp: [%8ld.%09ld]\n", evname, offset, ts->tv_sec, ts->tv_nsec);
 }

 

static void handle_event(struct mon_ctx *ctx, int event_type,
                         unsigned int line_offset,
                         const struct timespec *timestamp)
{
        if (!ctx->silent) {
                if (ctx->fmt)
                        event_print_custom(line_offset, timestamp,
                                           event_type, ctx);
                else
                        event_print_human_readable(line_offset,
                                                   timestamp, event_type);
        }

	printf("ads131m02: handle event \n");
        ctx->events_done++;
}
/*
static void handle_event(struct mon_ctx *ctx, int event_type, unsigned int line_offset, const struct timespec *timestamp)
{
         event_print_human_readable(line_offset, timestamp, event_type);
        printf("Interrupt Recived\n"); 
	interrupt_flag=1;
//       mode |= SPI_LOOP;      
         ctx->events_done++;
 
 }
*/

//Callback Function to call on event occurrence.
 static int event_callback(int event_type, unsigned int line_offset, const struct timespec *timestamp, void *data)
 {
         struct mon_ctx *ctx = data;
//         printf("%s %d \n",__func__,__LINE__);
	 printf("ads131m02: ADC event call back started \n");

	 switch (event_type)
         {
                 case GPIOD_CTXLESS_EVENT_CB_RISING_EDGE:
                 case GPIOD_CTXLESS_EVENT_CB_FALLING_EDGE:
                         handle_event(ctx, event_type, line_offset, timestamp);
                         break;
                 default:
                         /*
                          * REVISIT: This happening would indicate a problem in the library.
                          */
                          return GPIOD_CTXLESS_EVENT_CB_RET_OK;
         }
  //       printf("%s %d \n",__func__,__LINE__);
         printf("ads131m02: ADC event call back check1 \n");
         if ((ctx->events_wanted && ctx->events_done >= ctx->events_wanted) )
         {
//           printf("%s %d \n",__func__,__LINE__);
             printf("ads131m02: ADC event all back check2 \n");
             return GPIOD_CTXLESS_EVENT_CB_RET_STOP;
         }
  //       printf("%s %d \n",__func__,__LINE__);
         printf("ads131m02: ADC event call back check3 \n");
         return GPIOD_CTXLESS_EVENT_CB_RET_OK;
 }

static int poll_callback(unsigned int num_lines,
                         struct gpiod_ctxless_event_poll_fd *fds,
                         const struct timespec *timeout, void *data)
{
        struct pollfd pfds[GPIOD_LINE_BULK_MAX_LINES + 1];
        struct mon_ctx *ctx = data;
        int cnt, ts, rv;
        unsigned int i;

        for (i = 0; i < num_lines; i++) {
                pfds[i].fd = fds[i].fd;
                pfds[i].events = POLLIN | POLLPRI;
        }

	printf("ads131m02: ADC poll callback started");
        pfds[i].fd = ctx->sigfd;
        pfds[i].events = POLLIN | POLLPRI;

        ts = timeout->tv_sec * 1000 + timeout->tv_nsec / 1000000;

        cnt = poll(pfds, num_lines + 1, ts);
        if (cnt < 0)
                return GPIOD_CTXLESS_EVENT_POLL_RET_ERR;
        else if (cnt == 0)
                return GPIOD_CTXLESS_EVENT_POLL_RET_TIMEOUT;

        rv = cnt;
        for (i = 0; i < num_lines; i++) {
                if (pfds[i].revents) {
                        fds[i].event = true;
                        if (!--cnt)
                                return rv;
                }
        }

	//printf("ads131m02: ADC poll call back completed \n");
        /*
         * If we're here, then there's a signal pending. No need to read it,
         * we know we should quit now.
         */
        close(ctx->sigfd);

	 printf("ads131m02: ADC poll call back completed \n");

        return GPIOD_CTXLESS_EVENT_POLL_RET_STOP;
}

#if 0
int syncResetGPIOOp(int value) 
{
       int ret = -1;
       const struct mode_mapping *mode = &modes[MODE_EXIT];
       struct callback_data cbdata;
       bool active_low = false;
       bool active_high = true;
       int flags = GPIOD_CTXLESS_FLAG_BIAS_PULL_DOWN;
       int *values;
       unsigned int num_lines;
       unsigned int *offsets;

       num_lines = 1;
       values = malloc(sizeof(*values) * num_lines);
       offsets = malloc(sizeof(*offsets) * num_lines);

        offsets[0] = 15;
//	memset(value, 0, sizeof(*values) * num_lines );
        memset(&cbdata, 0, sizeof(cbdata));
        printf("ads131m02: Init GPIO started 22222 \n");

       //cbdata = NULL;
        ret = gpiod_ctxless_set_value_multiple_ext(
                "5", offsets, value,
                "1", active_high, "gpioset",
                mode->callback, &cbdata, flags);

	printf("check error %s\n", strerror(errno));
        printf("ads131m02: sync Reset GPIO \n");
	printf("return value of reset gpio %d \n",ret);
        printf("return value of reset gpio %x \n",ret);
	
	free(values);

        return ret;
}
#endif


int make_signalfd(void)
{
	sigset_t sigmask;
	int sigfd, rv;

	sigemptyset(&sigmask);
	sigaddset(&sigmask, SIGTERM);
	sigaddset(&sigmask, SIGINT);

	rv = sigprocmask(SIG_BLOCK, &sigmask, NULL);
	if (rv < 0)
		 // die("error masking signals: %s", strerror(errno));
		return 0;

	sigfd = signalfd(-1, &sigmask, 0);
	if (sigfd < 0)
		//die("error creating signalfd: %s", strerror(errno));
		return 0;

	return sigfd;
}
/*
void die(const char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
//	fprintf(stderr, "%s: ", program_invocation_name);
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	va_end(va);

	exit(EXIT_FAILURE);
}
*/
/*
const char *get_progname(void)
{
	return program_invocation_name;
}
*/
int syncResetGPIOOp(int value) 
{
	func("started sync RESET GPIO func \n");
	struct gpiohandle_request req;
        struct gpiohandle_data data;
        char chrdev_name[20];
        int fd, ret;

        strcpy(chrdev_name, "/dev/gpiochip5");
	//printf("reset ADC started");

        /*  Open device: gpiochip0 for GPIO bank A */
        fd = open(chrdev_name, 0);
        if (fd == -1) {
                ret = -errno;
                fprintf(stderr, "Failed to open %s\n", chrdev_name);
	//printf("reset ADC gpio bank set");
                return ret;
        }

	//printf("reset ADC gpio bank set1111");

        /* request GPIO line: GPIO_A_14 */
        req.lineoffsets[0] = 15;
        req.flags = GPIOHANDLE_REQUEST_OUTPUT;
        memcpy(req.default_values, &data, sizeof(req.default_values));
        strcpy(req.consumer_label, "ADC_reset_pin");
        req.lines  = 1;

	//printf("reset ADC gpio pin set");
        ret = ioctl(fd, GPIO_GET_LINEHANDLE_IOCTL, &req);
        if (ret == -1) {
                ret = -errno;
                fprintf(stderr, "Failed to issue GET LINEHANDLE IOCTL (%d)\n",
                        ret);
        }
        if (close(fd) == -1)
                perror("Failed to close GPIO character device file");

        /*  Start led blinking 
        while(1) {

                data.values[0] = !data.values[0];
                ret = ioctl(req.fd, GPIOHANDLE_SET_LINE_VALUES_IOCTL, &data);
                if (ret == -1) {
                        ret = -errno;
                        fprintf(stderr, "Failed to issue %s (%d)\n",
                                        "GPIOHANDLE_SET_LINE_VALUES_IOCTL", ret);
                }
                sleep(1);
        }

        printf("reset ADC led blinking"); */

        /*  release line */
        ret = close(req.fd);
        if (ret == -1) {
                perror("Failed to close GPIO LINEHANDLE device file");
                ret = -errno;
        }
	printf("ADC reset gpio set\n");
        return ret;
}


/*
int DRDYGPIOop()
{
     func("started sync RESET GPIO func \n");
     printf("DRDY GPIO started");
     int ret, event_type;
     int flags = 0;
     bool active_low = false;
     struct mon_ctx ctx;
     memset(&ctx, 0, sizeof(ctx));
     struct timespec timeout = {10, 0};
     int num_lines;

     event_type = GPIOD_CTXLESS_EVENT_BOTH_EDGES;
     flags = GPIOD_CTXLESS_FLAG_BIAS_PULL_DOWN; 

     num_lines = 1;
     ret =  gpiod_ctxless_event_monitor_multiple_ext(
                        DRDY_DEVICE, event_type, DRDY_OFFSETS, num_lines,
                        active_low, "gpiomon", &timeout, poll_callback,
                        GPIO_DRDY_IRQHandler, &ctx, flags);

     printf("ads131m02: DRDY GPIO \n");
     printf("return value of DRDY gpio %d \n",ret);
     printf("return value of DRDY gpio %x \n",ret);

     return ret;
}
*/


int DRDYGPIOop()
{
	unsigned int offsets[GPIOD_LINE_BULK_MAX_LINES], num_lines = 0, offset;
	bool active_low = false, watch_rising = false, watch_falling = true;
	int flags = 0;
	int rv, i, event_type;
	struct timespec timeout = {10, 0};
	struct mon_ctx ctx;
        interrupt_flag = 0;
	memset(&ctx, 0, sizeof(ctx));

	int ret = 0;
	int fd;
	uint32_t request;
{
	//printf("gpiomon main start\n");
	active_low = false;
        //printf("active low = false\n");
        flags = GPIOD_CTXLESS_FLAG_BIAS_PULL_DOWN;

	if (watch_rising && !watch_falling)
                 {
                         event_type = GPIOD_CTXLESS_EVENT_RISING_EDGE;
                         printf("event is at rising edge\n");
                 }
 
                 else if (watch_falling && !watch_rising)
                 {
                         event_type = GPIOD_CTXLESS_EVENT_FALLING_EDGE;
                         //printf("event is at falling edge\n");
                 }
 
                 else
                 {
                         event_type = GPIOD_CTXLESS_EVENT_BOTH_EDGES;
                         printf("event is at both edgesi\n");
                 }
 
                 offsets[0] = 12;
		num_lines = 1;
 //                char* device = "gpiochip4 4";
 
                 ctx.sigfd = make_signalfd();
 
                 rv =  gpiod_ctxless_event_monitor_multiple_ext(
                       "5", event_type, offsets, num_lines,
                       active_low, "gpiomon", &timeout, poll_callback,
                       event_callback, &ctx, flags);
                 //printf("gpio5 12 is set\n");
         }
return ret;
}


//****************************************************************************
//
// GPIO initialization
//
//****************************************************************************



//*****************************************************************************
//
//! Configures the MCU's GPIO pins that interface with the ADC.
//!
//! \fn void InitGPIO(void)
//!
//! \return None.
//
//*****************************************************************************


/* Do we need to handle the all below GPIO through is function or SPI will handlw it */
/* gpio mon */
//#define SYNC_RESET_DEVICE     ("2")
//#define SYNC_RESET_OFFSETS     ("21")    

//#define DRDY_DEVICE     ("5") // BANK
//#define DRDY_OFFSETS     ("4")    // PIN

void InitGPIO(void)
{
    int ret = -1;
    int value =0;
    //printf("\n ads131m02: Init GPIO started 11111 \n ");

    ret = syncResetGPIOOp(value);

    //printf("\n RESET GPIO set with ret value %d \n",ret);

    ret = DRDYGPIOop();

    //printf("\n DRDY GPIO set with ret value %d \n",ret);



    /* --- INSERT YOUR CODE HERE --- */
    // NOTE: Not all hardware implementations may control each of these pins...

    /* Enable the clock to the GPIO Port K and wait for it to be ready 
    MAP_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOK);
    while(!(SysCtlPeripheralReady(SYSCTL_PERIPH_GPIOK)))
    {
    } */

 /* Configure the GPIO for 'nSYNC_nRESET' as output and set high */
    //MAP_GPIOPinTypeGPIOOutput(nSYNC_nRESET_PORT, nSYNC_nRESET_PIN);
    //MAP_GPIOPinWrite(nSYNC_nRESET_PORT, nSYNC_nRESET_PIN, nSYNC_nRESET_PIN);
/*    int syncResetGPIOOp(int value) {
        int ret = -1;
        const struct mode_mapping *mode = &modes[MODE_EXIT];
        struct callback_data cbdata;
        bool active_low = false;
        int flags = GPIOD_CTXLESS_FLAG_BIAS_PULL_DOWN | GPIOD_CTXLESS_FLAG_OPEN_SOURCE;
        
	printf("ads131m02: Init GPIO started 22222 ");

        //cbdata = NULL;
        ret = gpiod_ctxless_set_value_multiple_ext(
                SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, value,
                1, active_low, "gpioset",
                mode->callback, &cbdata, flags);

	printf("ads131m02: sync Reset GPIO");

        return ret;
        }


    // Configure the GPIO for 'nCS' as output and set high 
    //MAP_GPIOPinTypeGPIOOutput(nCS_PORT, nCS_PIN);
    //MAP_GPIOPinWrite(nCS_PORT, nCS_PIN, nCS_PIN);

    //Enable the clock to the GPIO Port M and wait for it to be ready 
    //MAP_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOM);
    //while(!(SysCtlPeripheralReady(SYSCTL_PERIPH_GPIOM)))
    //{
    //}

    // Configure the GPIO for 'nDRDY' as input with falling edge interrupt 
    // GPIOIntRegister(nDRDY_PORT, GPIO_DRDY_IRQHandler);
    //MAP_GPIOPinTypeGPIOInput(nDRDY_PORT, nDRDY_PIN);
    //MAP_GPIOIntTypeSet(nDRDY_PORT, nDRDY_PIN, GPIO_FALLING_EDGE);
    //MAP_GPIOIntEnable(nDRDY_PORT, nDRDY_PIN);
    //MAP_IntEnable(nDRDY_INT);
    int DRDYGPIOop()
    {
    int ret, event_type;
    bool active_low = false;
    struct mon_ctx ctx;
    memset(&ctx, 0, sizeof(ctx));
    struct timespec timeout = {10, 0};

    event_type = GPIOD_CTXLESS_EVENT_BOTH_EDGES;

    ret =  gpiod_ctxless_event_monitor_multiple_ext(
                        DRDY_DEVICE, event_type, DRDY_OFFSETS, 1,
                        active_low, "gpiomon", &timeout, poll_callback,
                        GPIO_DRDY_IRQHandler, &ctx, GPIOD_CTXLESS_FLAG_BIAS_PULL_DOWN);
    
    printf("ads131m02: DRDY GPIO ");
    return ret;
   } */


}

//*****************************************************************************
//
// Interrupt handler for nDRDY GPIO
//
//*****************************************************************************



//*****************************************************************************
//
//! Interrupt handler for /DRDY falling edge interrupt.
//!
//! \fn void GPIO_DRDY_IRQHandler(void)
//!
//! \return None.
//
//*****************************************************************************

/* Do we need to handle DRDY gpio like handled in processor code */
/* gpio mon  call bacl reg */

void GPIO_DRDY_IRQHandler(void)
{
    /* --- INSERT YOUR CODE HERE --- */
    //NOTE: You many need to rename or register this interrupt function for your processor

    // Possible ways to handle this interrupt:
    // If you decide to read data here, you may want to disable other interrupts to avoid partial data reads.

    // In this example we set a flag and exit the interrupt routine. In the main program loop, your application can examine
    // all state flags and decide which state (operation) to perform next.

    /* Get the interrupt status from the GPIO and clear the status */
    //uint32_t getIntStatus = MAP_GPIOIntStatus(nDRDY_PORT, true);



    /* Check if the nDRDY pin triggered the interrupt */
   //if(getIntStatus & nDRDY_PIN)
   // {
        /* Interrupt action: Set a flag */
     //   flag_nDRDY_INTERRUPT = true;
   // }


    /* Clear interrupt gpio set function for clear the pin */
    //MAP_GPIOIntClear(nDRDY_PORT, getIntStatus);

       //int clear_DRDY(int value) {
        int ret = -1;
        const struct mode_mapping *mode = &modes[MODE_EXIT];
        struct callback_data cbdata;
        int flags = 0;
        bool active_low = false;
        //cbdata = NULL;
        ret = gpiod_ctxless_set_value_multiple_ext(
                SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, 0,
                1, active_low, "gpioset",
                mode->callback, &cbdata, flags);

	printf("ads131m02: ADC GPIO_DRDY_IRQHandler \n");

        usleep(3);
        return ret;
}

    // NOTE: We add a short delay at the end to prevent re-entrance. Refer to E2E issue:
    // https://e2e.ti.com/support/microcontrollers/tiva_arm/f/908/p/332605/1786938#1786938
    //SysCtlDelay(3);
   // udelay(3);

//****************************************************************************
//
// GPIO helper functions
//
//****************************************************************************



//*****************************************************************************
//
//! Reads that current state of the /CS GPIO pin.
//!
//! \fn bool getCS(void)
//!
//! \return boolean ('true' if /CS is high, 'false' if /CS is low).
//
//*****************************************************************************

/* Do not need to handle the CS pin with code spi will handle the CS pin*/
/*
bool getCS(void)
{
    /* --- INSERT YOUR CODE HERE --- */
 //   return (bool) GPIOPinRead(nCS_PORT, nCS_PIN);
//} 



//*****************************************************************************
//
//! Reads that current state of the nSYNC/nRESET GPIO pin.
//!
//! \fn bool getSYNC_RESET(void)
//!
//! \return boolean ('true' if nSYNC/nRESET is high, 'false' if nSYNC/nRESET is low).
//
//*****************************************************************************
/* Do we need to handle the RESET pin or spi will handle the RESET pin like CS pin.....gpio mon */
/* gpioget */
bool getSYNC_RESET(void)
{
    int ret = -1;
    int flags = 0;
    int *values;
    bool active_low = false;
    /* --- INSERT YOUR CODE HERE --- 
    return (bool) GPIOPinRead(nSYNC_nRESET_PORT, nSYNC_nRESET_PIN); */
    bool SYNC_RESET_get(void) {
    ret = gpiod_ctxless_get_value_multiple_ext(SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, values,
                                                  1, active_low,
                                                  "gpioget", flags);

    printf("ads131m02: ADC get sync reset \n");
    return ret;
        }
}

//*****************************************************************************
//
//! Controls the state of the /CS GPIO pin.
//!
//! \fn void setCS(const bool state)
//!
//! \param state boolean indicating which state to set the /CS pin (0=low, 1=high)
//!
//! NOTE: The 'HIGH' and 'LOW' macros defined in hal.h can be passed to this
//! function for the 'state' parameter value.
//!
//! \return None.
//
//*****************************************************************************

/* Not need to handle, SPI will handle that CS pin functionality */
/*
void setCS(const bool state)
{
    /* --- INSERT YOUR CODE HERE --- */

    // td(CSSC) delay
    //if(state) { SysCtlDelay(2); }

   // uint8_t value = (uint8_t) (state ? nCS_PIN : 0);
    //MAP_GPIOPinWrite(nCS_PORT, nCS_PIN, value);

    // td(SCCS) delay
  //  if(!state) { SysCtlDelay(2); }
//}



//*****************************************************************************
//
//! Controls the state of the nSYNC/nRESET GPIO pin.
//!
//! \fn void setSYNC_RESET(const bool state)
//!
//! \param state boolean indicating which state to set the nSYNC/nRESET pin (0=low, 1=high)
//!
//! NOTE: The 'HIGH' and 'LOW' macros defined in hal.h can be passed to this
//! function for the 'state' parameter value.
//!
//! \return None.
//
//*****************************************************************************
// gpioset same function use
void setSYNC_RESET(const bool state)
{
    /* --- INSERT YOUR CODE HERE --- */
    //uint8_t value = (uint8_t) (state ? nSYNC_nRESET_PIN : 0);
    //MAP_GPIOPinWrite(nSYNC_nRESET_PORT, nSYNC_nRESET_PIN, value);
    int ret = -1;
    const struct mode_mapping *mode = &modes[MODE_EXIT];
    int *values,flags = 0;
    bool active_low = false;
    struct callback_data cbdata;

    ret = gpiod_ctxless_set_value_multiple_ext(
                SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, 1,
                1, active_low, "gpioset",
                mode->callback, &cbdata, flags);

    //printf("ads131m02: ADC set SYNC reset \n");
    return ret;

}


//*****************************************************************************
//
//! Toggles the "nSYNC/nRESET" pin to trigger a synchronization
//! (LOW, delay 2 us, then HIGH).
//!
//! \fn void toggleSYNC(void)
//!
//! \return None.
//
//*****************************************************************************
//gpio set function use
void toggleSYNC(void)
{
    /* --- INSERT YOUR CODE HERE --- */
    //MAP_GPIOPinWrite(nSYNC_nRESET_PORT, nSYNC_nRESET_PIN, 0);
    int ret = -1;
    const struct mode_mapping *mode = &modes[MODE_EXIT];
    int *values,flags = 0;
    bool active_low = false;
    struct callback_data cbdata;


    ret = gpiod_ctxless_set_value_multiple_ext(
                SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, 0,
                1, active_low, "gpioset",
                mode->callback, &cbdata, flags);

    printf("ads131m02: ADC toggle sync 1 \n");
    // nSYNC pulse width must be between 1 and 2,048 CLKIN periods
    usleep(2);

    ret = gpiod_ctxless_set_value_multiple_ext(
                SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, 1,
                1, active_low, "gpioset",
                mode->callback, &cbdata, flags);

    printf("ads131m02: ADC toggle syc 2 \n");
    return ret;
}

//*****************************************************************************
//
//! Toggles the "nSYNC/nRESET" pin to trigger a reset
//! (LOW, delay 2 ms, then HIGH).
//!
//! \fn void toggleRESET(void)
//!
//! \return None.
//
//*****************************************************************************
//gpio set same function
void toggleRESET(void)
{
    /* --- INSERT YOUR CODE HERE --- */
    //MAP_GPIOPinWrite(nSYNC_nRESET_PORT, nSYNC_nRESET_PIN, 0);
    int ret = -1;
    const struct mode_mapping *mode = &modes[MODE_EXIT];
    int *values,flags = 0;
    bool active_low = false;
    struct callback_data cbdata;

    //printf("ads131m02: ADC toggle reset 1 \n");
    
    ret = gpiod_ctxless_set_value_multiple_ext(
                SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, 0,
                1, active_low, "gpioset",
                mode->callback, &cbdata, flags);

    //printf("ads131m02: ADC toggle reset 2 \n");
    // Minimum /RESET pulse width (tSRLRST) equals 2,048 CLKIN periods (1 ms @ 2.048 MHz)
    usleep(2*1000);

    //MAP_GPIOPinWrite(nSYNC_nRESET_PORT, nSYNC_nRESET_PIN, nSYNC_nRESET_PIN);

    ret = gpiod_ctxless_set_value_multiple_ext(
                SYNC_RESET_DEVICE, SYNC_RESET_OFFSETS, 1,
                1, active_low, "gpioset",
                mode->callback, &cbdata, flags);
    
    //printf("ads131m02: ADC toggle syc 3 \n");

    // tREGACQ delay before communicating with the device again
     usleep(5);

    // NOTE: The ADS131M0x's next response word should be (0xFF20 | CHANCNT).
    // A different response may be an indication that the device did not reset.

    // Update register array
    restoreRegisterDefaults();

    //printf("ads131m02: ADC toggle syc 4 \n");
    // Write to MODE register to enforce mode settings
    writeSingleRegister(MODE_ADDRESS, MODE_DEFAULT);
    //printf("ads131m02: ADC toggle syc 4 wr single reg \n");
}

//*****************************************************************************
//
//! Waits for the nDRDY interrupt or until the specified timeout occurs.
//!
//! \fn bool waitForDRDYinterrupt(const uint32_t timeout_ms)
//!
//! \param timeout_ms number of milliseconds to wait before timeout event.
//!
//! \return Returns 'true' if nDRDY interrupt occurred before the timeout.
//
//*****************************************************************************

bool waitForDRDYinterrupt(const uint32_t timeout_ms)
{
    /* --- INSERT YOUR CODE HERE ---
     * Poll the nDRDY GPIO pin until it goes low. To avoid potential infinite
     * loops, you may also want to implement a timer interrupt to occur after
     * the specified timeout period, in case the nDRDY pin is not active.
     * Return a boolean to indicate if nDRDY went low or if a timeout occurred.
     */

    // Convert ms to a # of loop iterations, OR even better use a timer here...
    uint32_t timeout = timeout_ms * 6000;   // convert to # of loop iterations
    printf("ads131m02: wait for drdy interrupt \n");
    // Reset interrupt flag
    flag_nDRDY_INTERRUPT = false;

    printf("ads131m02: ADC drdy flag check\n");
    // Enable interrupts
    //IntMasterEnable();

    // Wait for nDRDY interrupt or timeout - each iteration is about 20 ticks
    do {
        timeout--;
    } while (!flag_nDRDY_INTERRUPT && (timeout > 0));
 // Reset interrupt flag
    flag_nDRDY_INTERRUPT = false;

    // Timeout counter greater than zero indicates that an interrupt occurred
    return (timeout > 0);
}


//****************************************************************************
//
// SPI Communication
//
//****************************************************************************

#define SSI_BASE_ADDR       (SSI3_BASE)

//*****************************************************************************
//
//! Configures the MCU's SPI peripheral, for interfacing with the ADC.
//!
//! \fn void InitSPI(void)
//!
//! \return None.
//
//*****************************************************************************
void InitSPI(void)
{
    /* --- INSERT YOUR CODE HERE ---
     * NOTE: The ADS131M0x operates in SPI mode 1 (CPOL = 0, CPHA = 1).
     */

     int32_t ret;
        uint8_t nRetryCount = 0U;

	//printf("ads131m02: ADC init SPI \n");
        while (nRetryCount < ADC_RETRY_CNT)
        {
                //printf("ads131m02: ADC starting init spi \n");
		fd = open(ADC_SPI_DEVICE_NODE, O_RDWR);
                if (fd < 0)
                {
			nRetryCount++;
                        continue;
                }
                else
                {
                        /* spi mode */
                        ret = ioctl(fd, (u_int64_t)SPI_IOC_WR_MODE32, &mode);
                        if (ret != 0)
                        {
			 	nRetryCount++;
                                continue;
                        }

                        ret = ioctl(fd, (u_int64_t)SPI_IOC_RD_MODE32, &mode);
                        if (ret != 0)
                        {
                                nRetryCount++;
                                continue;
                        }

                        /* bits per word */
                        ret = ioctl(fd, (u_int64_t)SPI_IOC_WR_BITS_PER_WORD, &bits);
                        if (ret != 0)
                        {
                                nRetryCount++;
                                continue;
                        }

			ret = ioctl(fd, (u_int64_t)SPI_IOC_RD_BITS_PER_WORD, &bits);
                        if (ret != 0)
                        {
                                nRetryCount++;
                                continue;
                        }

                        /* max speed hz */
                        ret = ioctl(fd, (u_int64_t)SPI_IOC_WR_MAX_SPEED_HZ, &speed);
                        if (ret != 0)
                        {
                                nRetryCount++;
                                continue;
                        }
                        ret = ioctl(fd, (u_int64_t)SPI_IOC_RD_MAX_SPEED_HZ, &speed);
                        if (ret != 0)
                        {
                                nRetryCount++;
                                continue;
                        }
                }
                break;
        }
	printf("spi mode: 0x%x\n", mode);
	printf("bits per word: %u\n", bits);
	printf("max speed: %u Hz (%u kHz)\n", speed, speed/1000);
	//printf("Init GPIO completed\n");
        return -1;

}


static void spiSendReceiveArrays(int fd_t, uint8_t const *tx, uint8_t const *rx, size_t len)
{
        int ret;
        int out_fd;
        static char *output_file;
        static int verbose;
        static uint32_t mode;
        static uint8_t bits = 8;
        static uint32_t speed = 20000000;
        static uint16_t delay;
        struct spi_ioc_transfer tr = {
                .tx_buf = (unsigned long)tx,
                .rx_buf = (unsigned long)rx,
                .len = len,
                .delay_usecs = delay,
                .speed_hz = speed,
                .bits_per_word = bits,
        };

	//printf("ads131m02: ADC started with spiSendReceiveArrays \n");

        if (mode & SPI_TX_OCTAL)
                tr.tx_nbits = 8;
        else if (mode & SPI_TX_QUAD)
                tr.tx_nbits = 4;
        else if (mode & SPI_TX_DUAL)
                tr.tx_nbits = 2;
        if (mode & SPI_RX_OCTAL)
                tr.rx_nbits = 8;
        else if (mode & SPI_RX_QUAD)
                tr.rx_nbits = 4;
        else if (mode & SPI_RX_DUAL)
                tr.rx_nbits = 2;
        if (!(mode & SPI_LOOP)) {
                if (mode & (SPI_TX_OCTAL | SPI_TX_QUAD | SPI_TX_DUAL))
                        tr.rx_buf = 0;
                else if (mode & (SPI_RX_OCTAL | SPI_RX_QUAD | SPI_RX_DUAL))
                        tr.tx_buf = 0;
        }

//	tr.tx_nbits = 2;
//	tr.rx_nbits = 2;
	//printf("ads131m02: fd_t %d \n",fd_t);
	//printf("ads131m02: fd value %d \n",fd);
	//printf("ads131m02: tx value %d \n",tr.tx_nbits);
	//printf("ads131m02: rx value %d \n",tr.rx_nbits);


       	ret = ioctl(fd, SPI_IOC_MESSAGE(1), &tr);
        if (ret < 1) {
		printf("ads131m02: fd_t %d \n",fd_t);
		printf("ads131m02: %d \n",fd);
       	        printf("ads131m02: %d \n",tr.tx_nbits);
        	printf("ads131m02: %d \n",tr.rx_nbits);
	
                printf("can't send spi message \n");
		
		return -1;
	}

        if (verbose)
                hex_dump(tx, len, 32, "TX");

        if (output_file) {
                out_fd = open(output_file, O_WRONLY | O_CREAT | O_TRUNC, 0666);
                if (out_fd < 0)
                        printf("could not open output file");
			return -1;
                ret = write(out_fd, rx, len);
                if (ret != len)
                        printf("not all bytes written to output file");
			return -1;
                close(out_fd);
        }

        if (verbose)
                hex_dump(rx, len, 32, "RX");
}





//*****************************************************************************
//
//! Getter function to access registerMap array from outside of this module.
//!
//! \fn uint16_t getRegisterValue(uint8_t address)
//!
//! NOTE: The internal registerMap arrays stores the last know register value,
//! since the last read or write operation to that register. This function
//! does not communicate with the device to retrieve the current register value.
//! For the most up-to-date register data or retrieving the value of a hardware
//! controlled register, it is recommend to use readSingleRegister() to read the
//! current register value.
//!
//! \return unsigned 16-bit register value.
//
//*****************************************************************************
uint16_t getRegisterValue(uint8_t address)
{
    assert(address < NUM_REGISTERS);
    return registerMap[address];
}



//*****************************************************************************
//
//! Example start up sequence for the ADS131M0x.
//!
//! \fn void adcStartup(void)
//!
//! Before calling this function, the device must be powered,
//! the SPI/GPIO pins of the MCU must have already been configured,
//! and (if applicable) the external clock source should be provided to CLKIN.
//!
//! \return None.
//
//*****************************************************************************
void adcStartup(void)
{
	printf("ads131m02: ADC startup \n");
	/* (OPTIONAL) Provide additional delay time for power supply settling */
	delay_ms(50);

	/* (REQUIRED) Set nRESET pin high for ADC operation */
	setSYNC_RESET(HIGH);
	
	//printf("RESET pin is HIGH");
	/* (OPTIONAL) Toggle nRESET pin to ensure default register settings. */
	/* NOTE: This also ensures that the device registers are unlocked.	 */
	toggleRESET();

	printf("toggle the RESET pin\n");

    /* (REQUIRED) Initialize internal 'registerMap' array with device default settings */
	restoreRegisterDefaults();
	printf("Restore Reg Defaults\n");


    /* (OPTIONAL) Validate first response word when beginning SPI communication: (0xFF20 | CHANCNT) */
	uint16_t response = sendCommand(OPCODE_NULL);

	/* (OPTIONAL) Define your initial register settings here */
	writeSingleRegister(CLOCK_ADDRESS, (CLOCK_DEFAULT & ~CLOCK_OSR_MASK) | CLOCK_OSR_256);

        printf("Initial REG setting\n");

    /* (REQUIRED) Configure MODE register settings
     * NOTE: This function call is required here for this particular code implementation to work.
     * This function will enforce the MODE register settings as selected in the 'ads131m0x.h' header file.
     */
	writeSingleRegister(MODE_ADDRESS, MODE_DEFAULT);

	printf("mode reg setting\n");

    /* (OPTIONAL) Read back all registers */

	/* (OPTIONAL) Check STATUS register for faults */
	printf("Startup ADC function completed\n");
	printf("printing data first time\n ");
	readData(&DataStruct);
        sleep(30);
	printf("printing data second time\n ");
	readData(&DataStruct);
	sleep(30);
	printf("printing data third time\n ");
	readData(&DataStruct);

}



//*****************************************************************************
//
//! Reads the contents of a single register at the specified address.
//!
//! \fn uint16_t readSingleRegister(uint8_t address)
//!
//! \param address is the 8-bit address of the register to read.
//!
//! \return Returns the 8-bit register read result.
//
//*****************************************************************************
uint16_t readSingleRegister(uint8_t address)
{
	//printf("ads131m02: ADC read single reg \n");
	/* Check that the register address is in range */
	assert(address < NUM_REGISTERS);

// Build TX and RX byte array
#ifdef ENABLE_CRC_IN
    uint8_t dataTx[8] = { 0 };      // 2 words, up to 4 bytes each = 8 bytes maximum
    uint8_t dataRx[8] = { 0 };
#else
    uint8_t dataTx[4] = { 0 };      // 1 word, up to 4 bytes long = 4 bytes maximum
    uint8_t dataRx[4] = { 0 };
#endif
    uint16_t opcode = OPCODE_RREG | (((uint16_t) address) << 7);
    uint8_t numberOfBytes = buildSPIarray(&opcode, 1, dataTx);
     
	// [FRAME 1] Send RREG command
//	spiSendReceiveArrays(dataTx, dataRx, numberOfBytes);
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);
	// [FRAME 2] Send NULL command to retrieve the register data
	registerMap[address] = sendCommand(OPCODE_NULL);

	return registerMap[address];
}



//*****************************************************************************
//
//! Writes data to a single register.
//!
//! \fn void writeSingleRegister(uint8_t address, uint16_t data)
//!
//! \param address is the address of the register to write to.
//! \param data is the value to write.
//!
//! This command will be ignored if device registers are locked.
//!
//! \return None.
//
//*****************************************************************************
void writeSingleRegister(uint8_t address, uint16_t data)
{
    //printf("ads131m02: ADC write single reg \n");
    /* Check that the register address is in range */
    assert(address < NUM_REGISTERS);

    // (OPTIONAL) Enforce certain register field values when
    // writing to the MODE register to fix the operation mode
    if (MODE_ADDRESS == address)
    {
        data = enforce_selected_device_modes(data);
    }

    // Build TX and RX byte array
#ifdef ENABLE_CRC_IN
    uint8_t dataTx[12] = { 0 };     // 3 words, up to 4 bytes each = 12 bytes maximum
    uint8_t dataRx[12] = { 0 };
#else
    uint8_t dataTx[8] = { 0 };      // 2 words, up to 4 bytes long = 8 bytes maximum
    uint8_t dataRx[8] = { 0 };
#endif
    uint16_t opcodes[2];
    opcodes[0] = OPCODE_WREG | (((uint16_t) address) << 7);
    opcodes[1] = data;
   // uint8_t numberOfBytes = buildSPIarray(&opcodes[0], 2, dataTx);
    uint8_t numberOfBytes = 2;

    // Send command
    //spiSendReceiveArrays(dataTx, dataRx, numberOfBytes);
    spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);

    // Update internal array
    registerMap[address] = data;

    // (RECOMMENDED) Read back register to confirm register write was successful
    readSingleRegister(address);

    // NOTE: Enabling the CRC words in the SPI command will NOT prevent an invalid W
}



//*****************************************************************************
//
//! Reads ADC data.
//!
//! \fn bool readData(adc_channel_data *DataStruct)
//!
//! \param *DataStruct points to an adc_channel_data type-defined structure/
//!
//! NOTE: Should be called after /DRDY goes low, and not during a /DRDY falling edge!
//!
//! \return Returns true if the CRC-OUT of the data read detects an error.
//
//*****************************************************************************
bool readData(adc_channel_data *DataStruct)
{
    int i;
    uint8_t crcTx[4]                        = { 0 };
    uint8_t dataRx[4]                       = { 0 };
    uint8_t dataTx[4]                       = { 0 };
    uint8_t bytesPerWord                    = getWordByteLength();

    printf("ads131m02: ADC read data\n");
#ifdef ENABLE_CRC_IN
    // Build CRC word (only if "RX_CRC_EN" register bit is enabled)
    uint16_t crcWordIn = calculateCRC(&DataTx[0], bytesPerWord * 2, 0xFFFF);
    crcTx[0] = upperByte(crcWordIn);
    crcTx[1] = lowerByte(crcWordIn);
#endif
    bool b_unlock_error;
    uint16_t opcode = OPCODE_UNLOCK;
    uint8_t numberOfBytes = buildSPIarray(&opcode, 1, dataTx);
    /* Set the nCS pin LOW */
    //setCS(LOW);

    // Send NULL word, receive response word
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
        spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->response = combineBytes(dataRx[0], dataRx[1]);

    // (OPTIONAL) Do something with the response (STATUS) word.
    // ...Here we only use the response for calculating the CRC-OUT
    //uint16_t crcWord = calculateCRC(&dataRx[0], bytesPerWord, 0xFFFF);

    // (OPTIONAL) Ignore CRC error checking
    uint16_t crcWord = 0;

    // Send 2nd word, receive channel 1 data
    for (i = 0; i < bytesPerWord; i++)
    {
       // dataRx[i] = spiSendReceiveByte(crcTx[i]);
	dataRx[i] = crcTx[i];
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);
    }
    DataStruct->channel0 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);
    printf("channel0 value is %d in dec \n", DataStruct->channel0);
    printf("channel0 value is %x in hex \n", DataStruct->channel0);
    sleep(2);
 
    /*
    while(1) {
	    printf("channel0 value is %d", DataStruct->channel0);
	sleep(2);
    }

    */
#if (CHANNEL_COUNT > 1)

    // Send 3rd word, receive channel 2 data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->channel1 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);
    printf("channel1 value is %d in dec \n", DataStruct->channel1);
    printf("channel1 value is %x in hex \n", DataStruct->channel1);
    sleep(2);

#endif
#if (CHANNEL_COUNT > 2)

    // Send 4th word, receive channel 3 data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->channel2 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);
#endif
#if (CHANNEL_COUNT > 3)

    // Send 5th word, receive channel 4 data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->channel3 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);

#endif
#if (CHANNEL_COUNT > 4)

    // Send 6th word, receive channel 5 data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->channel4 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);

#endif
#if (CHANNEL_COUNT > 5)

    // Send 7th word, receive channel 6 data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->channel5 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);

#endif
#if (CHANNEL_COUNT > 6)

    // Send 8th word, receive channel 7 data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
        spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->channel6 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);

#endif
#if (CHANNEL_COUNT > 7)

    // Send 9th word, receive channel 8 data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->channel7 = signExtend(&dataRx[0]);
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);

#endif

    // Send the next word, receive CRC data
    for (i = 0; i < bytesPerWord; i++)
    {
        dataRx[i] = 0x00;
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx,1);
    }
    DataStruct->crc = combineBytes(dataRx[0], dataRx[1]);

    /* NOTE: If we continue calculating the CRC with a matching CRC, the result should be zero.
     * Any non-zero result will indicate a mismatch.
     */
    //crcWord = calculateCRC(&dataRx[0], bytesPerWord, crcWord);

    /* Set the nCS pin HIGH */
    //setCS(HIGH);

    // Returns true when a CRC error occurs
    return ((bool) crcWord);
}



//*****************************************************************************
//
//! Sends the specified SPI command to the ADC (NULL, STANDBY, or WAKEUP).
//!
//! \fn uint16_t sendCommand(uint16_t opcode)
//!
//! \param opcode SPI command byte.
//!
//! NOTE: Other commands have their own dedicated functions to support
//! additional functionality.
//!
//! \return ADC response byte (typically the STATUS byte).
//
//*****************************************************************************
uint16_t sendCommand(uint16_t opcode)
{
    //printf("ads131m02: ADC send command");
    /* Assert if this function is used to send any of the following opcodes */
    assert(OPCODE_RREG != opcode);      /* Use "readSingleRegister()"   */
    assert(OPCODE_WREG != opcode);      /* Use "writeSingleRegister()"  */
    assert(OPCODE_LOCK != opcode);      /* Use "lockRegisters()"        */
    assert(OPCODE_UNLOCK != opcode);    /* Use "unlockRegisters()"      */
    assert(OPCODE_RESET != opcode);     /* Use "resetDevice()"          */

    // Build TX and RX byte array
#ifdef ENABLE_CRC_IN
    uint8_t dataTx[8] = { 0 };      // 2 words, up to 4 bytes each = 8 bytes maximum
    uint8_t dataRx[8] = { 0 };
#else
    uint8_t dataTx[4] = { 0 };      // 1 word, up to 4 bytes long = 4 bytes maximum
    uint8_t dataRx[4] = { 0 };
#endif
    uint8_t numberOfBytes = buildSPIarray(&opcode, 1, dataTx);

    /* Set the nCS pin LOW */
    //setCS(LOW);

    // Send the opcode (and crc word, if enabled)
    int i;
    for (i = 0; i < numberOfBytes; i++)
    {
       //dataRx[i] = spiSendReceiveByte(dataTx[i]);
	dataRx[i] = dataTx[i];
        spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);
    }

    /* Set the nCS pin HIGH */
    //setCS(HIGH);

    // Combine response bytes and return as a 16-bit word
    uint16_t adcResponse = combineBytes(dataRx[0], dataRx[1]);
    return adcResponse;
}



//*****************************************************************************
//
//! Resets the device.
//!
//! \fn void resetDevice(void)
//!
//! NOTE: This function does not capture DOUT data, but it could be modified
//! to do so.
//!
//! \return None.
//
//*****************************************************************************
void resetDevice(void)
{
    printf("ads131m02: ADC reset device ");
    // Build TX and RX byte array
#ifdef ENABLE_CRC_IN
    uint8_t dataTx[8] = { 0 };      // 2 words, up to 4 bytes each = 8 bytes maximum
    uint8_t dataRx[8] = { 0 };    // Only needed if capturing data
#else
    uint8_t dataTx[4] = { 0 };      // 1 word, up to 4 bytes long = 4 bytes maximum
    uint8_t dataRx[4] = { 0 };    // Only needed if capturing data
#endif
    uint16_t opcode         = OPCODE_RESET;
    uint8_t numberOfBytes   = buildSPIarray(&opcode, 1, dataTx);

    uint8_t bytesPerWord    = wlength_byte_values[WLENGTH];
    uint8_t wordsInFrame    = CHANNEL_COUNT + 2;

    // Set the nCS pin LOW
    //setCS(LOW);

    // Send the opcode (and CRC word, if enabled)
    int i;
    for (i = 0; i < numberOfBytes; i++)
    {
         //spiSendReceiveByte(dataTx[i]);
	 spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);
    }

    // Finish sending remaining bytes
    for (i = numberOfBytes; i < (wordsInFrame * bytesPerWord); i++)
    {
        //spiSendReceiveByte(0x00);
	spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);
    }

    // NOTE: The ADS131M0x's next response word should be (0xFF20 | CHANCNT),
    // if the response is 0x0011 (acknowledge of RESET command), then the device
    // did not receive a full SPI frame and the reset did not occur!

    // Set the nCS pin HIGH
   // setCS(HIGH);

    // tSRLRST delay, ~1ms with 2.048 MHz fCLK
    delay_ms(1);

    // Update register setting array to keep software in sync with device
    restoreRegisterDefaults();

    // Write to MODE register to enforce mode settings
    writeSingleRegister(MODE_ADDRESS, MODE_DEFAULT);
}



//*****************************************************************************
//
//! Sends the LOCK command and verifies that registers are locked.
//!
//! \fn bool lockRegisters(void)
//!
//! \return boolean to indicate if an error occurred (0 = no error; 1 = error)
//
//*****************************************************************************
bool lockRegisters(void)
{
   printf("ads131m02: ADC lock Reg");
    bool b_lock_error;

    // Build TX and RX byte array
#ifdef ENABLE_CRC_IN
    uint8_t dataTx[8] = { 0 };      // 2 words, up to 4 bytes each = 8 bytes maximum
    uint8_t dataRx[8] = { 0 };
#else
    uint8_t dataTx[4] = { 0 };      // 1 word, up to 4 bytes long = 4 bytes maximum
    uint8_t dataRx[4] = { 0 };
#endif
    uint16_t opcode         = OPCODE_LOCK;
    uint8_t numberOfBytes   = buildSPIarray(&opcode, 1, dataTx);

    // Send command
    //spiSendReceiveArrays(dataTx, dataRx, numberOfBytes);
    spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);

    /* (OPTIONAL) Check for SPI errors by sending the NULL command and checking STATUS */

    /* (OPTIONAL) Read back the STATUS register and check if LOCK bit is set... */
    readSingleRegister(STATUS_ADDRESS);
    if (!SPI_LOCKED) { b_lock_error = true; }

    /* If the STATUS register is NOT read back,
     * then make sure to manually update the global register map variable... */
    //registerMap[STATUS_ADDRESS]  |= STATUS_LOCK_LOCKED;

    /* (OPTIONAL) Error handler */
    if (b_lock_error)
    {
        // Insert error handler function call here...
    }

    return b_lock_error;
}



//*****************************************************************************
//
//! Sends the UNLOCK command and verifies that registers are unlocked
//!
//! \fn bool unlockRegisters(void)
//!
//! \return boolean to indicate if an error occurred (0 = no error; 1 = error)
//
//*****************************************************************************
bool unlockRegisters(void)
{
	printf("ads131m02: ADC unlock reg");
	bool b_unlock_error;

    // Build TX and RX byte array
#ifdef ENABLE_CRC_IN
    uint8_t dataTx[8] = { 0 };      // 2 words, up to 4 bytes each = 8 bytes maximum
    uint8_t dataRx[8] = { 0 };
#else
    uint8_t dataTx[4] = { 0 };      // 1 word, up to 4 bytes long = 4 bytes maximum
    uint8_t dataRx[4] = { 0 };
#endif
    uint16_t opcode = OPCODE_UNLOCK;
    uint8_t numberOfBytes = buildSPIarray(&opcode, 1, dataTx);

    // Send command
//    spiSendReceiveArrays(dataTx, dataRx, numberOfBytes);
    spiSendReceiveArrays(fd, (uint8_t const *)dataTx, (uint8_t const *)dataRx, numberOfBytes);

    /* (OPTIONAL) Check for SPI errors by sending the NULL command and checking STATUS */

    /* (OPTIONAL) Read the STATUS register and check if LOCK bit is cleared... */
    readSingleRegister(STATUS_ADDRESS);
    if (SPI_LOCKED) { b_unlock_error = true; }

    /* If the STATUS register is NOT read back,
     * then make sure to manually update the global register map variable... */
    //registerMap[STATUS_ADDRESS]  &= !STATUS_LOCK_LOCKED;

    /* (OPTIONAL) Error handler */
    if (b_unlock_error)
    {
        // Insert error handler function call here...
    }

    return b_unlock_error;
}



//*****************************************************************************
//
//! Calculates the 16-bit CRC for the selected CRC polynomial.
//!
//! \fn uint16_t calculateCRC(const uint8_t dataBytes[], uint8_t numberBytes, uint16_t initialValue)
//!
//! \param dataBytes[] pointer to first element in the data byte array
//! \param numberBytes number of bytes to be used in CRC calculation
//! \param initialValue the seed value (or partial crc calculation), use 0xFFFF when beginning a new CRC computation
//!
//! NOTE: This calculation is shown as an example and is not optimized for speed.
//!
//! \return 16-bit calculated CRC word
//
//*****************************************************************************
uint16_t calculateCRC(const uint8_t dataBytes[], uint8_t numberBytes, uint16_t initialValue)
{
	printf("ads131m02: ADC calculate CRC");
	/* Check that "dataBytes" is not a null pointer */
	assert(dataBytes != 0x00);

	int         bitIndex, byteIndex;
	bool        dataMSb;						/* Most significant bit of data byte */
	bool        crcMSb;						    /* Most significant bit of crc byte  */
	//uint8_t     bytesPerWord = wlength_byte_values[WLENGTH];

	/*
     * Initial value of crc register
     * NOTE: The ADS131M0x defaults to 0xFFFF,
     * but can be set at function call to continue an on-going calculation
     */
    uint16_t crc = initialValue;

    #ifdef CRC_CCITT
    /* CCITT CRC polynomial = x^16 + x^12 + x^5 + 1 */
    const uint16_t poly = 0x1021;
    #endif

    #ifdef CRC_ANSI
    /* ANSI CRC polynomial = x^16 + x^15 + x^2 + 1 */
    const uint16_t poly = 0x8005;
    #endif

    //
    // CRC algorithm
    //

    // Loop through all bytes in the dataBytes[] array
	for (byteIndex = 0; byteIndex < numberBytes; byteIndex++)
	{
	    // Point to MSb in byte
	    bitIndex = 0x80u;

	    // Loop through all bits in the current byte
	    while (bitIndex > 0)
	    {
	        // Check MSB's of data and crc
	        dataMSb = (bool) (dataBytes[byteIndex] & bitIndex);
	        crcMSb  = (bool) (crc & 0x8000u);

	        crc <<= 1;              /* Left shift CRC register */

	        // Check if XOR operation of MSBs results in additional XOR operations
	        if (dataMSb ^ crcMSb)
	        {
	            crc ^= poly;        /* XOR crc with polynomial */
	        }

	        /* Shift MSb pointer to the next data bit */
	        bitIndex >>= 1;
	    }
	}

	return crc;
}



//*****************************************************************************
//
//! Updates the registerMap[] array to its default values.
//!
//! \fn void restoreRegisterDefaults(void)
//!
//! NOTES:
//! - If the MCU keeps a copy of the ADS131M0x register settings in memory,
//! then it is important to ensure that these values remain in sync with the
//! actual hardware settings. In order to help facilitate this, this function
//! should be called after powering up or resetting the device (either by
//! hardware pin control or SPI software command).
//!
//! - Reading back all of the registers after resetting the device can
//! accomplish the same result; however, this might be problematic if the
//! device was previously in CRC mode or the WLENGTH was modified, since
//! resetting the device exits these modes. If the MCU is not aware of this
//! mode change, then read register commands will return invalid data due to
//! the expectation of data appearing in a different byte position.
//!
//! \return None.
//
//*****************************************************************************
void restoreRegisterDefaults(void)
{
    registerMap[ID_ADDRESS]             =   0x00;               /* NOTE: This a read-only register */
    registerMap[STATUS_ADDRESS]         =   STATUS_DEFAULT;
    registerMap[MODE_ADDRESS]           =   MODE_DEFAULT;
    registerMap[CLOCK_ADDRESS]          =   CLOCK_DEFAULT;
    registerMap[GAIN1_ADDRESS]          =   GAIN1_DEFAULT;
    registerMap[GAIN2_ADDRESS]          =   GAIN2_DEFAULT;
   //registerMap[GAIN1_ADDRESS]          =   0x11;
   // registerMap[GAIN2_ADDRESS]          =   0x11;

    registerMap[CFG_ADDRESS]            =   CFG_DEFAULT;
    registerMap[THRSHLD_MSB_ADDRESS]    =   THRSHLD_MSB_DEFAULT;
    registerMap[THRSHLD_LSB_ADDRESS]    =   THRSHLD_LSB_DEFAULT;
    registerMap[CH0_CFG_ADDRESS]        =   CH0_CFG_DEFAULT;
    registerMap[CH0_OCAL_MSB_ADDRESS]   =   CH0_OCAL_MSB_DEFAULT;
    registerMap[CH0_OCAL_LSB_ADDRESS]   =   CH0_OCAL_LSB_DEFAULT;
    registerMap[CH0_GCAL_MSB_ADDRESS]   =   CH0_GCAL_MSB_DEFAULT;
    registerMap[CH0_GCAL_LSB_ADDRESS]   =   CH0_GCAL_LSB_DEFAULT;
#if (CHANNEL_COUNT > 1)
    registerMap[CH1_CFG_ADDRESS]        =   CH1_CFG_DEFAULT;
    registerMap[CH1_OCAL_MSB_ADDRESS]   =   CH1_OCAL_MSB_DEFAULT;
    registerMap[CH1_OCAL_LSB_ADDRESS]   =   CH1_OCAL_LSB_DEFAULT;
    registerMap[CH1_GCAL_MSB_ADDRESS]   =   CH1_GCAL_MSB_DEFAULT;
    registerMap[CH1_GCAL_LSB_ADDRESS]   =   CH1_GCAL_LSB_DEFAULT;
#endif
#if (CHANNEL_COUNT > 2)
    registerMap[CH2_CFG_ADDRESS]        =   CH2_CFG_DEFAULT;
    registerMap[CH2_OCAL_MSB_ADDRESS]   =   CH2_OCAL_MSB_DEFAULT;
    registerMap[CH2_OCAL_LSB_ADDRESS]   =   CH2_OCAL_LSB_DEFAULT;
    registerMap[CH2_GCAL_MSB_ADDRESS]   =   CH2_GCAL_MSB_DEFAULT;
    registerMap[CH2_GCAL_LSB_ADDRESS]   =   CH2_GCAL_LSB_DEFAULT;
#endif
#if (CHANNEL_COUNT > 3)
    registerMap[CH3_CFG_ADDRESS]        =   CH3_CFG_DEFAULT;
    registerMap[CH3_OCAL_MSB_ADDRESS]   =   CH3_OCAL_MSB_DEFAULT;
    registerMap[CH3_OCAL_LSB_ADDRESS]   =   CH3_OCAL_LSB_DEFAULT;
    registerMap[CH3_GCAL_MSB_ADDRESS]   =   CH3_GCAL_MSB_DEFAULT;
    registerMap[CH3_GCAL_LSB_ADDRESS]   =   CH3_GCAL_LSB_DEFAULT;
#endif
#if (CHANNEL_COUNT > 4)
    registerMap[CH4_CFG_ADDRESS]        =   CH4_CFG_DEFAULT;
    registerMap[CH4_OCAL_MSB_ADDRESS]   =   CH4_OCAL_MSB_DEFAULT;
    registerMap[CH4_OCAL_LSB_ADDRESS]   =   CH4_OCAL_LSB_DEFAULT;
    registerMap[CH4_GCAL_MSB_ADDRESS]   =   CH4_GCAL_MSB_DEFAULT;
    registerMap[CH4_GCAL_LSB_ADDRESS]   =   CH4_GCAL_LSB_DEFAULT;
#endif
#if (CHANNEL_COUNT > 5)
    registerMap[CH5_CFG_ADDRESS]        =   CH5_CFG_DEFAULT;
    registerMap[CH5_OCAL_MSB_ADDRESS]   =   CH5_OCAL_MSB_DEFAULT;
    registerMap[CH5_OCAL_LSB_ADDRESS]   =   CH5_OCAL_LSB_DEFAULT;
    registerMap[CH5_GCAL_MSB_ADDRESS]   =   CH5_GCAL_MSB_DEFAULT;
    registerMap[CH5_GCAL_LSB_ADDRESS]   =   CH5_GCAL_LSB_DEFAULT;
#endif
#if (CHANNEL_COUNT > 6)
    registerMap[CH6_CFG_ADDRESS]        =   CH6_CFG_DEFAULT;
    registerMap[CH6_OCAL_MSB_ADDRESS]   =   CH6_OCAL_MSB_DEFAULT;
    registerMap[CH6_OCAL_LSB_ADDRESS]   =   CH6_OCAL_LSB_DEFAULT;
    registerMap[CH6_GCAL_MSB_ADDRESS]   =   CH6_GCAL_MSB_DEFAULT;
    registerMap[CH6_GCAL_LSB_ADDRESS]   =   CH6_GCAL_LSB_DEFAULT;
#endif
#if (CHANNEL_COUNT > 7)
    registerMap[CH7_CFG_ADDRESS]        =   CH7_CFG_DEFAULT;
    registerMap[CH7_OCAL_MSB_ADDRESS]   =   CH7_OCAL_MSB_DEFAULT;
    registerMap[CH7_OCAL_LSB_ADDRESS]   =   CH7_OCAL_LSB_DEFAULT;
    registerMap[CH7_GCAL_MSB_ADDRESS]   =   CH7_GCAL_MSB_DEFAULT;
    registerMap[CH7_GCAL_LSB_ADDRESS]   =   CH7_GCAL_LSB_DEFAULT;
#endif
    registerMap[REGMAP_CRC_ADDRESS]     =   REGMAP_CRC_DEFAULT;
}



//****************************************************************************
//
// Helper functions
//
//****************************************************************************


//*****************************************************************************
//
//! Takes a 16-bit word and returns the most-significant byte.
//!
//! \fn uint8_t upperByte(uint16_t uint16_Word)
//!
//! \param temp_word is the original 16-bit word.
//!
//! \return 8-bit most-significant byte.
//
//*****************************************************************************
uint8_t upperByte(uint16_t uint16_Word)
{
    uint8_t msByte;
    msByte = (uint8_t) ((uint16_Word >> 8) & 0x00FF);

    return msByte;
}



//*****************************************************************************
//
//! Takes a 16-bit word and returns the least-significant byte.
//!
//! \fn uint8_t lowerByte(uint16_t uint16_Word)
//!
//! \param temp_word is the original 16-bit word.
//!
//! \return 8-bit least-significant byte.
//
//*****************************************************************************
uint8_t lowerByte(uint16_t uint16_Word)
{
    uint8_t lsByte;
    lsByte = (uint8_t) (uint16_Word & 0x00FF);

    return lsByte;
}



//*****************************************************************************
//
//! Takes two 8-bit words and returns a concatenated 16-bit word.
//!
//! \fn uint16_t combineBytes(uint8_t upperByte, uint8_t lowerByte)
//!
//! \param upperByte is the 8-bit value that will become the MSB of the 16-bit word.
//! \param lowerByte is the 8-bit value that will become the LSB of the 16-bit word.
//!
//! \return concatenated 16-bit word.
//
//*****************************************************************************
uint16_t combineBytes(uint8_t upperByte, uint8_t lowerByte)
{
    uint16_t combinedValue;
    combinedValue = ((uint16_t) upperByte << 8) | ((uint16_t) lowerByte);

    return combinedValue;
}



//*****************************************************************************
//
//! Combines ADC data bytes into a single signed 32-bit word.
//!
//! \fn int32_t combineDataBytes(const uint8_t dataBytes[])
//!
//! \param dataBytes is a pointer to uint8_t[] where the first element is the MSB.
//!
//! \return Returns the signed-extend 32-bit result.
//
//*****************************************************************************
int32_t signExtend(const uint8_t dataBytes[])
{

#ifdef WORD_LENGTH_24BIT

    int32_t upperByte   = ((int32_t) dataBytes[0] << 24);
    int32_t middleByte  = ((int32_t) dataBytes[1] << 16);
    int32_t lowerByte   = ((int32_t) dataBytes[2] << 8);

    return (((int32_t) (upperByte | middleByte | lowerByte)) >> 8);     // Right-shift of signed data maintains signed bit

#elif defined WORD_LENGTH_32BIT_SIGN_EXTEND

    int32_t signByte    = ((int32_t) dataBytes[0] << 24);
    int32_t upperByte   = ((int32_t) dataBytes[1] << 16);
    int32_t middleByte  = ((int32_t) dataBytes[2] << 8);
    int32_t lowerByte   = ((int32_t) dataBytes[3] << 0);

    return (signByte | upperByte | middleByte | lowerByte);

#elif defined WORD_LENGTH_32BIT_ZERO_PADDED

    int32_t upperByte   = ((int32_t) dataBytes[0] << 24);
    int32_t middleByte  = ((int32_t) dataBytes[1] << 16);
    int32_t lowerByte   = ((int32_t) dataBytes[2] << 8);

    return (((int32_t) (upperByte | middleByte | lowerByte)) >> 8);     // Right-shift of signed data maintains signed bit

#elif defined WORD_LENGTH_16BIT_TRUNCATED

    int32_t upperByte   = ((int32_t) dataBytes[0] << 24);
    int32_t lowerByte   = ((int32_t) dataBytes[1] << 16);

    return (((int32_t) (upperByte | lowerByte)) >> 16);                 // Right-shift of signed data maintains signed bit

#endif
}



//****************************************************************************
//
// Internal functions
//
//****************************************************************************


//*****************************************************************************
//
//! Builds SPI TX data arrays according to number of opcodes provided and
//! currently programmed device word length.
//!
//! \fn uint8_t buildSPIarray(const uint16_t opcodeArray[], uint8_t numberOpcodes, uint8_t byteArray[])
//!
//! \param opcodeArray[] pointer to an array of 16-bit opcodes to use in the SPI command.
//! \param numberOpcodes the number of opcodes provided in opcodeArray[].
//! \param byteArray[] pointer to an array of 8-bit SPI bytes to send to the device.
//!
//! NOTE: The calling function must ensure it reserves sufficient memory for byteArray[]!
//!
//! \return number of bytes added to byteArray[].
//
//*****************************************************************************


uint8_t buildSPIarray(const uint16_t opcodeArray[], uint8_t numberOpcodes, uint8_t byteArray[])
{
    /*
     * Frame size = opcode word(s) + optional CRC word
     * Number of bytes per word = 2, 3, or 4
     * Total bytes = bytes per word * number of words
     */
    uint8_t numberWords     = numberOpcodes + (SPI_CRC_ENABLED ? 1 : 0);
    uint8_t bytesPerWord    = getWordByteLength();
    uint8_t numberOfBytes   = numberWords * bytesPerWord;

    int i;
    for (i = 0; i < numberOpcodes; i++)
    {
        // NOTE: Be careful not to accidentally overflow the array here.
        // The array and opcodes are defined in the calling function, so
        // we are trusting that no mistakes were made in the calling function!
        byteArray[(i*bytesPerWord) + 0] = upperByte(opcodeArray[i]);
        byteArray[(i*bytesPerWord) + 1] = lowerByte(opcodeArray[i]);
    }

#ifdef ENABLE_CRC_IN
    // Calculate CRC and put it into TX array
    uint16_t crcWord = calculateCRC(&byteArray[0], numberOfBytes, 0xFFFF);
    byteArray[(i*bytesPerWord) + 0] = upperByte(crcWord);
    byteArray[(i*bytesPerWord) + 1] = lowerByte(crcWord);
#endif

    return numberOfBytes;
}



//*****************************************************************************
//
//! Modifies MODE register data to maintain device operation according to
//! preselected mode(s) (RX_CRC_EN, WLENGTH, etc.).
//!
//! \fn uint16_t enforce_selected_device_mode(uint16_t data)
//!
//! \param data uint16_t register data.
//!
//! \return uint16_t modified register data.
//
//*****************************************************************************
uint16_t enforce_selected_device_modes(uint16_t data)
{


    ///////////////////////////////////////////////////////////////////////////
    // Enforce RX_CRC_EN setting

#ifdef ENABLE_CRC_IN
    // When writing to the MODE register, ensure RX_CRC_EN bit is ALWAYS set
    data |= MODE_RX_CRC_EN_ENABLED;
#else
    // When writing to the MODE register, ensure RX_CRC_EN bit is NEVER set
    data &= ~MODE_RX_CRC_EN_ENABLED;
#endif // ENABLE_CRC_IN


    ///////////////////////////////////////////////////////////////////////////
    // Enforce WLENGH setting

#ifdef WORD_LENGTH_24BIT
    // When writing to the MODE register, ensure WLENGTH bits are ALWAYS set to 01b
    data = (data & ~MODE_WLENGTH_MASK) | MODE_WLENGTH_24BIT;
#elif defined WORD_LENGTH_32BIT_SIGN_EXTEND
    // When writing to the MODE register, ensure WLENGH bits are ALWAYS set to 11b
    data = (data & ~MODE_WLENGTH_MASK) | MODE_WLENGTH_32BIT_MSB_SIGN_EXT;
#elif defined WORD_LENGTH_32BIT_ZERO_PADDED
    // When writing to the MODE register, ensure WLENGH bits are ALWAYS set to 10b
    data = (data & ~MODE_WLENGTH_MASK) | MODE_WLENGTH_32BIT_LSB_ZEROES;
#elif defined WORD_LENGTH_16BIT_TRUNCATED
    // When writing to the MODE register, ensure WLENGH bits are ALWAYS set to 00b
    data = (data & ~MODE_WLENGTH_MASK) | MODE_WLENGTH_16BIT;
#endif


    ///////////////////////////////////////////////////////////////////////////
    // Enforce DRDY_FMT setting

#ifdef DRDY_FMT_PULSE
    // When writing to the MODE register, ensure DRDY_FMT bit is ALWAYS set
    data = (data & ~MODE_DRDY_FMT_MASK) | MODE_DRDY_FMT_NEG_PULSE_FIXED_WIDTH;
#else
    // When writing to the MODE register, ensure DRDY_FMT bit is NEVER set
    data = (data & ~MODE_DRDY_FMT_MASK) | MODE_DRDY_FMT_LOGIC_LOW;
#endif


    ///////////////////////////////////////////////////////////////////////////
    // Enforce CRC_TYPE setting

#ifdef CRC_CCITT
    // When writing to the MODE register, ensure CRC_TYPE bit is NEVER set
    data = (data & ~STATUS_CRC_TYPE_MASK) | STATUS_CRC_TYPE_16BIT_CCITT;
#elif defined CRC_ANSI
    // When writing to the MODE register, ensure CRC_TYPE bit is ALWAYS set
    data = (data & ~STATUS_CRC_TYPE_MASK) | STATUS_CRC_TYPE_16BIT_ANSI;
#endif

    // Return modified register data
    return data;
}



//*****************************************************************************
//
//! Returns the ADS131M0x configured word length used for SPI communication.
//!
//! \fn uint8_t getWordByteLength(void)
//!
//! NOTE: It is important that the MODE register value stored in registerMap[]
//! remains in sync with the device. If these values get out of sync then SPI
//! communication may fail!
//!
//! \return SPI word byte length (2, 3, or 4)
//
//*****************************************************************************
uint8_t getWordByteLength(void)
{
    return wlength_byte_values[WLENGTH];
}


